@@ -23,11 +23,14 @@ from operator import itemgetter
 import functools
 import time
 
+
 from botocore.exceptions import NoCredentialsError
-from botocore.utils import normalize_url_path
+from botocore.utils import normalize_url_path, percent_encode_sequence
+from botocore.utils import percent_encode
 from botocore.compat import HTTPHeaders
-from botocore.compat import quote, unquote, urlsplit, parse_qs, urlencode
+from botocore.compat import quote, unquote, urlsplit, parse_qs
 from botocore.compat import urlunsplit
+from botocore.compat import encodebytes
 
 logger = logging.getLogger(__name__)
 
@@ -118,7 +121,7 @@ class SigV3Auth(BaseSigner):
         new_hmac = hmac.new(self.credentials.secret_key.encode('utf-8'),
                             digestmod=sha256)
         new_hmac.update(request.headers['Date'].encode('utf-8'))
-        encoded_signature = base64.encodestring(new_hmac.digest()).strip()
+        encoded_signature = encodebytes(new_hmac.digest()).strip()
         signature = ('AWS3-HTTPS AWSAccessKeyId=%s,Algorithm=%s,Signature=%s' %
                      (self.credentials.access_key, 'HmacSHA256',
                       encoded_signature.decode('utf-8')))
@@ -254,7 +257,8 @@ class SigV4Auth(BaseSigner):
         return '\n'.join(cr)
 
     def _normalize_url_path(self, path):
-        return normalize_url_path(path)
+        normalized_path = quote(normalize_url_path(path), safe='/~')
+        return normalized_path
 
     def scope(self, args):
         scope = [self.credentials.access_key]
@@ -378,7 +382,7 @@ class SigV4QueryAuth(SigV4Auth):
         # You can't mix the two types of params together, i.e just keep doing
         # new_query_params.update(op_params)
         # new_query_params.update(auth_params)
-        # urlencode(new_query_params)
+        # percent_encode_sequence(new_query_params)
         operation_params = ''
         if request.data:
             # We also need to move the body params into the query string.
@@ -388,8 +392,9 @@ class SigV4QueryAuth(SigV4Auth):
             query_dict.update(request.data)
             request.data = ''
         if query_dict:
-            operation_params = urlencode(query_dict) + '&'
-        new_query_string = operation_params + urlencode(auth_params)
+            operation_params = percent_encode_sequence(query_dict) + '&'
+        new_query_string = operation_params + \
+                percent_encode_sequence(auth_params)
         # url_parts is a tuple (and therefore immutable) so we need to create
         # a new url_parts with the new query string.
         # <part>   - <index>
@@ -452,7 +457,7 @@ class HmacV1Auth(BaseSigner):
         new_hmac = hmac.new(self.credentials.secret_key.encode('utf-8'),
                             digestmod=sha1)
         new_hmac.update(string_to_sign.encode('utf-8'))
-        return base64.encodestring(new_hmac.digest()).strip().decode('utf-8')
+        return encodebytes(new_hmac.digest()).strip().decode('utf-8')
 
     def canonical_standard_headers(self, headers):
         interesting_headers = ['content-md5', 'content-type', 'date']
